@@ -37,7 +37,7 @@ from . import calibration
 from .column_encoders import (CategoricalEncoder, ColumnEncoder,
                               NumericalEncoder, TfIdfEncoder)
 from .evaluation import evaluate_and_persist_metrics
-from .iterators import ImputerIterDf
+from .iterators import ImputerIterDf, INSTANCE_WEIGHT_COLUMN
 from .mxnet_input_symbols import Featurizer
 from .utils import (AccuracyMetric, ColumnOverwriteException,
                     LogMetricCallBack, MeanSymbol, get_context, logger,
@@ -69,7 +69,6 @@ class Imputer:
                  label_encoders: List[ColumnEncoder],
                  output_path="") -> None:
 
-
         self.ctx = None
         self.module = None
         self.data_encoders = data_encoders
@@ -129,7 +128,6 @@ class Imputer:
                         ", ".join(encoder_outputs), featurizer_type))
             # TODO: check whether encoder type matches requirements of featurizer
 
-
         # collect names of data and label columns
         input_col_names = [c.field_name for c in self.data_featurizers]
         label_col_names = list(itertools.chain(*[c.input_columns for c in self.label_encoders]))
@@ -166,12 +164,21 @@ class Imputer:
 
         """
 
-        logger.setLevel(level)
+        if os.access(os.path.dirname(filename), os.W_OK):
+            file_handler = FileHandler(filename, mode='a')
+            file_handler.setLevel(level)
+            file_handler.setFormatter(log_formatter)
+            logger.addHandler(file_handler)
+        else:
+            logger.warning("Could not attach file log handler, {} is not writable.".format(filename))
 
-        fileHandler = FileHandler(filename, mode='a')
-        fileHandler.setLevel(level)
-        fileHandler.setFormatter(log_formatter)
-        logger.addHandler(fileHandler)
+    def __close_filehandlers(self) -> None:
+        """Function to close connection with log file."""
+
+        handlers = logger.handlers[:]
+        for handler in handlers:
+            handler.close()
+            logger.removeHandler(handler)
 
     def __check_data(self, data_frame: pd.DataFrame) -> None:
         """
@@ -195,11 +202,12 @@ class Imputer:
             elif isinstance(col_enc, CategoricalEncoder):
                 values_not_in_test_set = set(col_enc.token_to_idx.keys()) - \
                                          set(data_frame[col_enc.input_columns[0]].unique())
-                logger.warning(
-                    "Test set does not contain any ocurrences of values [{}] in column [{}], "
-                    "consider using a more representative test set.".format(
-                        ", ".join(values_not_in_test_set),
-                        col_enc.input_columns[0]))
+                if len(values_not_in_test_set) > 0:
+                    logger.warning(
+                        "Test set does not contain any ocurrences of values [{}] in column [{}], "
+                        "consider using a more representative test set.".format(
+                            ", ".join(values_not_in_test_set),
+                            col_enc.input_columns[0]))
 
     def fit(self,
             train_df: pd.DataFrame,
@@ -244,7 +252,7 @@ class Imputer:
         self.final_fc_hidden_units = final_fc_hidden_units
 
         self.ctx = ctx
-        logger.info('Using [{}] as the context for training'.format(ctx))
+        logger.debug('Using [{}] as the context for training'.format(ctx))
 
         if (train_df is None) or (not isinstance(train_df, pd.core.frame.DataFrame)):
             raise ValueError("Need a non-empty DataFrame for fitting Imputer model")
@@ -256,7 +264,10 @@ class Imputer:
 
         self.__check_data(test_df)
 
-        self.module = self.__build_module(iter_train)
+        # to make consecutive calls to .fit() continue where the previous call finished
+        if self.module is None:
+            self.module = self.__build_module(iter_train)
+
         self.__fit_module(iter_train, iter_test, learning_rate, num_epochs, patience, weight_decay)
 
         # Check whether calibration is needed, if so ompute and set internal parameter
@@ -277,6 +288,8 @@ class Imputer:
         if self.is_explainable:
             self.__persist_class_prototypes(iter_train, train_df)
 
+        self.__close_filehandlers()
+
         return self
 
     def __persist_class_prototypes(self, iter_train, train_df):
@@ -288,7 +301,7 @@ class Imputer:
         """
 
         if len(self.label_encoders) > 1:
-            logger.warn('Persisting class prototypes works only for a single output column. '
+            logger.warning('Persisting class prototypes works only for a single output column. '
                         'Choosing ' + str(self.label_encoders[0].output_column) + '.')
         label_name = self.label_encoders[0].output_column
 
@@ -303,7 +316,7 @@ class Imputer:
         explainable_data_encoders_idx = []
         for encoder_idx, encoder in enumerate(self.data_encoders):
             if not (isinstance(encoder, TfIdfEncoder) or isinstance(encoder, CategoricalEncoder)):
-                logger.warn("Data encoder type {} incompatible for explaining classes".format(type(encoder)))
+                logger.warning("Data encoder type {} incompatible for explaining classes".format(type(encoder)))
             else:
                 explainable_data_encoders.append(encoder)
                 explainable_data_encoders_idx.append(encoder_idx)
@@ -329,7 +342,7 @@ class Imputer:
                         [np.sum(p_normalized[np.where(feature_matrix_scaled[0, :] == category)[0], :], axis=0)
                          for category in encoder.idx_to_token.keys()])))
             else:
-                logger.warn("column encoder not supported for explain.")
+                logger.warning("column encoder not supported for explain.")
 
         self.__class_patterns = class_patterns
 
@@ -452,7 +465,6 @@ class Imputer:
 
         return feature_dict
 
-
     def __fit_module(self,
                      iter_train: ImputerIterDf,
                      iter_test: ImputerIterDf,
@@ -508,7 +520,7 @@ class Imputer:
             except StopIteration:
                 # catch the StopIteration exception thrown when early stopping condition is reached
                 # this is ugly but the only way to use module api and have early stopping
-                logger.info("Stopping training, patience reached")
+                logger.debug("Stopping training, patience reached")
                 pass
 
         self.training_time = time.time() - start
@@ -552,7 +564,7 @@ class Imputer:
         for encoder in self.label_encoders:
             if not encoder.is_fitted():
                 encoder_type = type(encoder)
-                logger.info("Fitting label encoder {} on {} rows \
+                logger.debug("Fitting label encoder {} on {} rows \
                             of training data".format(encoder_type, len(train_df)))
                 encoder.fit(train_df)
 
@@ -565,18 +577,18 @@ class Imputer:
 
         test_df = self.__drop_missing_labels(test_df, how='all')
 
-        logger.info("Train: {}, Test: {}".format(len(train_df), len(test_df)))
+        logger.debug("Train: {}, Test: {}".format(len(train_df), len(test_df)))
 
         for encoder in self.data_encoders:
             if not encoder.is_fitted():
                 encoder_type = type(encoder)
-                logger.info(
-                    "Fitting data encoder {} on columns {} and {} rows of training data".format(
-                        encoder_type, ", ".join(encoder.input_columns), len(train_df)))
+                logger.debug(
+                    "Fitting data encoder {} on columns {} and {} rows of training data with parameters {}".format(
+                        encoder_type, ", ".join(encoder.input_columns), len(train_df), encoder.__dict__))
 
                 encoder.fit(train_df)
 
-        logger.info("Building Train Iterator with {} elements".format(len(train_df)))
+        logger.debug("Building Train Iterator with {} elements".format(len(train_df)))
         iter_train = ImputerIterDf(
             data_frame=train_df,
             data_columns=self.data_encoders,
@@ -584,7 +596,7 @@ class Imputer:
             batch_size=self.batch_size
         )
 
-        logger.info("Building Test Iterator with {} elements".format(len(test_df)))
+        logger.debug("Building Test Iterator with {} elements".format(len(test_df)))
         iter_test = ImputerIterDf(
             data_frame=test_df,
             data_columns=iter_train.data_columns,
@@ -649,7 +661,7 @@ class Imputer:
             if isinstance(col_enc, CategoricalEncoder):
                 predictions[att] = self.__filter_predictions(predictions[att], precision_threshold)
             else:
-                logger.info("Precision filtering only for CategoricalEncoder returning \
+                logger.debug("Precision filtering only for CategoricalEncoder returning \
                             {} unfiltered".format(att))
                 predictions[att] = predictions[att]
 
@@ -702,7 +714,7 @@ class Imputer:
                                           top_k_pred_idx]
                     top_k_predictions[att].append(label_proba_tuples)
             else:
-                logger.info(
+                logger.debug(
                     "Top-k only for CategoricalEncoder, dropping {}, {}".format(att, type(col_enc)))
                 top_k_predictions[att] = probas
 
@@ -817,14 +829,14 @@ class Imputer:
         predictions = self.predict_above_precision(data_frame, precision_threshold).items()
         for label, imputations in predictions:
             imputation_col = label + imputation_suffix
-            if data_frame.columns.contains(imputation_col):
+            if imputation_col in data_frame.columns:
                 raise ColumnOverwriteException(
                     "DataFrame contains column {}; remove column and try again".format(
                         imputation_col))
 
             if label not in numerical_outputs:
                 imputation_proba_col = label + score_suffix
-                if data_frame.columns.contains(imputation_proba_col):
+                if imputation_proba_col in data_frame.columns:
                     raise ColumnOverwriteException(
                         "DataFrame contains column {}; remove column and try again".format(
                             imputation_proba_col))
@@ -925,7 +937,7 @@ class Imputer:
                 # for NumericalEncoders, exclude rows that are nan
                 col_missing_idx = data_frame[col_enc.input_columns[0]].isna()
 
-            logger.info("Detected {} rows with missing labels \
+            logger.debug("Detected {} rows with missing labels \
                         for column {}".format(col_missing_idx.sum(), col_enc.input_columns[0]))
 
             if missing_idx == -1:
@@ -935,7 +947,7 @@ class Imputer:
             elif how == 'any':
                 missing_idx = missing_idx | col_missing_idx
 
-        logger.info("Dropping {}/{} rows".format(missing_idx.sum(), n_samples))
+        logger.debug("Dropping {}/{} rows".format(missing_idx.sum(), n_samples))
 
         return data_frame.loc[~missing_idx, :]
 
@@ -946,11 +958,11 @@ class Imputer:
 
         """
         best_model = glob.glob(self.module_path + "*{}.params".format(self.__get_best_epoch()))
-        logger.info("Keeping {}".format(best_model[0]))
+        logger.debug("Keeping {}".format(best_model[0]))
         worse_models = set(glob.glob(self.module_path + "*.params")) - set(best_model)
         # remove worse models
         for worse_epoch in worse_models:
-            logger.info("Deleting {}".format(worse_epoch))
+            logger.debug("Deleting {}".format(worse_epoch))
             os.remove(worse_epoch)
 
     def __get_best_epoch(self):
@@ -983,7 +995,7 @@ class Imputer:
 
         """
 
-        logger.info("Output path for loading Imputer {}".format(output_path))
+        logger.debug("Output path for loading Imputer {}".format(output_path))
         params = pickle.load(open(os.path.join(output_path, "imputer.pickle"), "rb"))
         imputer_signature = inspect.getfullargspec(Imputer.__init__)[0]
         # get constructor args
@@ -1000,15 +1012,23 @@ class Imputer:
         # the module path must be updated when loading the Imputer, too
         imputer.module_path = os.path.join(output_path, 'model')
         imputer.output_path = output_path
-        ctx = imputer.ctx
+        # make sure that the context for this deserialized model is available
+        ctx = get_context()
+
+        logger.debug("Loading mxnet model from {}".format(imputer.module_path))
+
+        # for categorical outputs, instance weight is added
+        if isinstance(imputer.label_encoders[0], NumericalEncoder):
+            data_names = [s.field_name for s in imputer.data_featurizers]
+        else:
+            data_names = [s.field_name for s in imputer.data_featurizers] + [INSTANCE_WEIGHT_COLUMN]
 
-        logger.info("Loading mxnet model from {}".format(imputer.module_path))
         # deserialize mxnet module
         imputer.module = mx.module.Module.load(
             imputer.module_path,
             imputer.__get_best_epoch(),
             context=ctx,
-            data_names=[s.field_name for s in imputer.data_featurizers],
+            data_names=data_names,
             label_names=[s.output_column for s in imputer.label_encoders]
         )
         return imputer
@@ -1058,12 +1078,12 @@ class Imputer:
         ece_pre = calibration.compute_ece(scores, labels)
         self.calibration_info['ece_pre'] = ece_pre
         self.calibration_info['reliability_pre'] = calibration.reliability(scores, labels)
-        logger.info('Expected calibration error: {:.1f}%'.format(100*ece_pre))
+        logger.debug('Expected calibration error: {:.1f}%'.format(100*ece_pre))
 
         temperature = calibration.fit_temperature(scores, labels)
         ece_post = calibration.compute_ece(scores, labels, temperature)
         self.calibration_info['ece_post'] = ece_post
-        logger.info('Expected calibration error after calibration: {:.1f}%'.format(100*ece_post))
+        logger.debug('Expected calibration error after calibration: {:.1f}%'.format(100*ece_post))
 
         # check whether calibration improves at all and apply
         if ece_pre - ece_post > 0:
@@ -1094,7 +1114,8 @@ class _MXNetModule:
         self.label_encoders = label_encoders
         self.final_fc_hidden_units = final_fc_hidden_units
 
-    def __call__(self, iter_train: ImputerIterDf) -> mx.mod.Module:
+    def __call__(self,
+                 iter_train: ImputerIterDf) -> mx.mod.Module:
         """
         Given a training iterator, build MXNet module and return it
 
@@ -1104,7 +1125,7 @@ class _MXNetModule:
 
         predictions, loss = self.__make_loss()
 
-        logger.info("Building output symbols")
+        logger.debug("Building output symbols")
         output_symbols = []
         for col_enc, output in zip(self.label_encoders, predictions):
             output_symbols.append(
@@ -1113,10 +1134,14 @@ class _MXNetModule:
         mod = mx.mod.Module(
             mx.sym.Group([loss] + output_symbols),
             context=self.ctx,
-            data_names=[name for name, dim in iter_train.provide_data],
+            # [name for name, dim in iter_train.provide_data],
+            data_names=[name for name, dim in iter_train.provide_data if name in loss.list_arguments()],
             label_names=[name for name, dim in iter_train.provide_label]
         )
-        mod.bind(data_shapes=iter_train.provide_data, label_shapes=iter_train.provide_label)
+
+        if mod.binded is False:
+            mod.bind(data_shapes=[d for d in iter_train.provide_data if d.name in loss.list_arguments()],  # iter_train.provide_data,
+                     label_shapes=iter_train.provide_label)
 
         return mod
 
@@ -1158,20 +1183,21 @@ class _MXNetModule:
                             data=latents,
                             num_hidden=layer)
 
+        instance_weight = mx.sym.Variable(INSTANCE_WEIGHT_COLUMN)
         pred = mx.sym.softmax(fully_connected)
         label = mx.sym.Variable(label_field_name)
 
         # assign to 0.0 the label values larger than number of classes so that they
         # do not contribute to the loss
 
-        logger.info("Building output of label {} with {} classes \
+        logger.debug("Building output of label {} with {} classes \
                      (including missing class)".format(label, num_labels))
 
         num_labels_vec = label * 0.0 + num_labels
         indices = mx.sym.broadcast_lesser(label, num_labels_vec)
         label = label * indices
 
-        # goes from (batch, 1) to (batch,) as it is required for softmax output
+        # goes from (batch, 1) to (batch,) as is required for softmax output
         label = mx.sym.split(label, axis=1, num_outputs=1, squeeze_axis=1)
 
         # mask entries when label is 0 (missing value)
@@ -1180,6 +1206,8 @@ class _MXNetModule:
 
         # compute the cross entropy only when labels are positive
         cross_entropy = mx.sym.pick(mx.sym.log_softmax(fully_connected), label) * -1 * positive_mask
+        # multiply loss by class weighting
+        cross_entropy = cross_entropy * mx.sym.pick(instance_weight, label)
 
         # normalize the cross entropy by the number of positive label
         num_positive_indices = mx.sym.sum(positive_mask)
@@ -1193,7 +1221,8 @@ class _MXNetModule:
         return pred, cross_entropy
 
     @staticmethod
-    def __make_numerical_loss(latents: mx.symbol, label_field_name: str) -> Tuple[Any, Any]:
+    def __make_numerical_loss(latents: mx.symbol,
+                              label_field_name: str) -> Tuple[Any, Any]:
         """
         Generate output symbol for univariate numeric loss
 
@@ -1217,7 +1246,7 @@ class _MXNetModule:
 
     def __make_loss(self, eps: float = 1e-5) -> Tuple[Any, Any]:
 
-        logger.info("Concatenating all {} latent symbols".format(len(self.data_featurizers)))
+        logger.debug("Concatenating all {} latent symbols".format(len(self.data_featurizers)))
 
         unique_input_field_names = set([feat.field_name for feat in self.data_featurizers])
         if len(unique_input_field_names) < len(self.data_featurizers):
@@ -1232,7 +1261,7 @@ class _MXNetModule:
         outputs = []
         for output_col in self.label_encoders:
             if isinstance(output_col, CategoricalEncoder):
-                logger.info("Constructing categorical loss for column {} and {} labels".format(
+                logger.debug("Constructing categorical loss for column {} and {} labels".format(
                     output_col.output_column, output_col.max_tokens))
                 outputs.append(
                     self.__make_categorical_loss(
@@ -1243,7 +1272,7 @@ class _MXNetModule:
                     )
                 )
             elif isinstance(output_col, NumericalEncoder):
-                logger.info(
+                logger.debug(
                     "Constructing numerical loss for column {}".format(output_col.output_column))
                 outputs.append(self.__make_numerical_loss(latents, output_col.output_column))
 
