@@ -17,24 +17,27 @@ DataWig SimpleImputer:
 Uses some simple default encoders and featurizers that usually yield decent imputation quality
 
 """
-import pickle
-import os
-import json
+import glob
 import inspect
+import json
+import os
+import pickle
+import shutil
 from typing import List, Dict, Any, Callable
-import itertools
+
 import mxnet as mx
 import pandas as pd
-import glob
-import shutil
+import numpy as np
 from pandas.api.types import is_numeric_dtype
-from sklearn.metrics import mean_squared_error, f1_score, precision_score, accuracy_score, recall_score
 
 from ._hpo import _HPO
-from .utils import logger, get_context, random_split, rand_string, flatten_dict, merge_dicts
+from .column_encoders import BowEncoder, CategoricalEncoder, NumericalEncoder, TfIdfEncoder
 from .imputer import Imputer
-from .column_encoders import BowEncoder, CategoricalEncoder, NumericalEncoder, ColumnEncoder, TfIdfEncoder
-from .mxnet_input_symbols import BowFeaturizer, NumericalFeaturizer, Featurizer, EmbeddingFeaturizer
+from .mxnet_input_symbols import BowFeaturizer, NumericalFeaturizer
+from .utils import logger, get_context, random_split, rand_string, flatten_dict, merge_dicts, set_stream_log_level
+from .imputer import Imputer
+from .iterators import INSTANCE_WEIGHT_COLUMN
+
 
 
 class SimpleImputer:
@@ -134,38 +137,38 @@ class SimpleImputer:
         self.string_columns = list(set(self.input_columns) - set(self.numeric_columns))
         self.output_type = 'numeric' if is_numeric_dtype(data_frame[self.output_column]) else 'string'
 
-        logger.info(
+        logger.debug(
             "Assuming {} numeric input columns: {}".format(len(self.numeric_columns),
                                                            ", ".join(self.numeric_columns)))
-        logger.info("Assuming {} string input columns: {}".format(len(self.string_columns),
+        logger.debug("Assuming {} string input columns: {}".format(len(self.string_columns),
                                                                   ", ".join(self.string_columns)))
 
     @staticmethod
     def _is_categorical(col: pd.Series,
                         n_samples: int = 100,
-                        min_value_histogram: float = 0.1) -> bool:
+                        max_unique_fraction=0.05) -> bool:
         """
 
         A heuristic to check whether a column is categorical:
-        a column is considered categorical (as opposed to a plain text column) if the least frequent value
-        occurs with a frequency of at least ``min_value_histogram``.
+        a column is considered categorical (as opposed to a plain text column)
+        if the relative cardinality is max_unique_fraction or less.
 
         :param col: pandas Series containing strings
         :param n_samples: number of samples used for heuristic (default: 100)
-        :min_value_histogram: minimum value in the normalized value histogram (default: 0.1)
+        :param max_unique_fraction: maximum relative cardinality.
 
         :return: True if the column is categorical according to the heuristic
 
         """
 
-        return col.sample(n=n_samples, replace=len(col) < n_samples).value_counts(
-            normalize=True).min() >= min_value_histogram
+        sample = col.sample(n=n_samples, replace=len(col) < n_samples).unique()
+
+        return sample.shape[0] / n_samples < max_unique_fraction
 
     def fit_hpo(self,
                 train_df: pd.DataFrame,
                 test_df: pd.DataFrame = None,
                 hps: dict = None,
-                strategy: str = 'random',
                 num_evals: int = 10,
                 max_running_hours: float = 96.0,
                 hpo_run_name: str = None,
@@ -175,8 +178,8 @@ class SimpleImputer:
                 test_split: float = .2,
                 weight_decay: List[float] = None,
                 batch_size: int = 16,
-                num_hash_bucket_candidates: List[float] = None,
-                tokens_candidates: List[str] = None,
+                num_hash_bucket_candidates: List[float] = [2 ** exp for exp in [12, 15, 18]],
+                tokens_candidates: List[str] = ['words', 'chars'],
                 numeric_latent_dim_candidates: List[int] = None,
                 numeric_hidden_layers_candidates: List[int] = None,
                 final_fc_hidden_units: List[List[int]] = None,
@@ -186,8 +189,7 @@ class SimpleImputer:
                 ctx: mx.context = get_context()) -> Any:
 
         """
-
-        Fits an imputer model with gridsearch hyperparameter optimization (hpo)
+        Fits an imputer model with hyperparameter optimization. The parameter ranges are searched randomly.
 
         Grids are specified using the *_candidates arguments (old)
         or with more flexibility via the dictionary hps.
@@ -199,7 +201,6 @@ class SimpleImputer:
             hps[column_name][parameter_name] is a list of parameter values for each input column.
             Further, hps[column_name]['type'] is in ['numeric', 'categorical', 'string'] and is
             inferred if not provided.
-        :param strategy: 'random' for random search or 'grid' for exhaustive search
         :param num_evals: number of evaluations for random search
         :param max_running_hours: Time before the hpo run is terminated in hours.
         :param hpo_run_name: string to identify the current hpo run.
@@ -230,55 +231,59 @@ class SimpleImputer:
             use is deprecated.
         :param ctx: List of mxnet contexts (if no gpu's available, defaults to [mx.cpu()])
             User can also pass in a list gpus to be used, ex. [mx.gpu(0), mx.gpu(2), mx.gpu(4)]
-            This parameter is deprecated.s
+            This parameter is deprecated.
 
         :return: pd.DataFrame with with hyper-parameter configurations and results
         """
 
         # generate dictionary with default hyperparameter settings. Overwrite these defaults
-        # with configurations that were passed via the old API where applicable.
+        # with configurations that were passed via this functions API wherever applicable.
         default_hps = dict()
         default_hps['global'] = dict()
-        default_hps['global']['learning_rate'] = \
-            learning_rate_candidates if learning_rate_candidates is not None else [1e-4, 1e-3]
-        default_hps['global']['weight_decay'] = \
-            weight_decay if weight_decay is not None else [0, 1e-4]
+        if learning_rate_candidates:
+            default_hps['global']['learning_rate'] = learning_rate_candidates
+        if weight_decay:
+            default_hps['global']['weight_decay'] = weight_decay
+        if num_epochs:
+            default_hps['global']['num_epochs'] = [num_epochs]
 
-        default_hps['global']['num_epochs'] = \
-            [num_epochs] if num_epochs is not None else [100]
-        default_hps['global']['patience'] = \
-            [patience] if patience is not None else [5]
-        default_hps['global']['batch_size'] = \
-            [batch_size] if batch_size is not None else [16]
-        default_hps['global']['final_fc_hidden_units'] = \
-            final_fc_hidden_units if final_fc_hidden_units is not None else [[], [100]]
-        default_hps['global']['concat_columns'] = [True, False]
+        if patience:
+            default_hps['global']['patience'] = [patience]
+
+        if batch_size:
+            default_hps['global']['batch_size'] = [batch_size]
+        if final_fc_hidden_units:
+            default_hps['global']['final_fc_hidden_units'] = final_fc_hidden_units
 
         default_hps['string'] = {}
-        default_hps['string']['max_tokens'] = \
-            num_hash_bucket_candidates if num_hash_bucket_candidates is not None else [2 ** 8]
-        default_hps['string']['tokens'] = \
-            [[cand] for cand in tokens_candidates] if tokens_candidates is not None else [['words']]
+        if num_hash_bucket_candidates:
+            default_hps['string']['max_tokens'] = num_hash_bucket_candidates
+
+        if tokens_candidates:
+            default_hps['string']['tokens'] = [[c] for c in tokens_candidates]
 
         default_hps['categorical'] = {}
-        default_hps['categorical']['max_tokens'] = \
-            num_hash_bucket_candidates if num_hash_bucket_candidates is not None else [2 ** 8]
-        default_hps['categorical']['embed_dim'] = [10]
+        if num_hash_bucket_candidates:
+            default_hps['categorical']['max_tokens'] = num_hash_bucket_candidates
 
         default_hps['numeric'] = {}
-        default_hps['numeric']['normalize'] = \
-            [normalize_numeric] if normalize_numeric is not None else [True]
-        default_hps['numeric']['numeric_latent_dim'] = \
-            numeric_latent_dim_candidates if numeric_latent_dim_candidates is not None else [10, 50]
-        default_hps['numeric']['numeric_hidden_layers'] = \
-            numeric_hidden_layers_candidates if numeric_hidden_layers_candidates is not None else [1, 2]
+        if normalize_numeric:
+            default_hps['numeric']['normalize'] = [normalize_numeric]
+        if numeric_latent_dim_candidates:
+            default_hps['numeric']['numeric_latent_dim'] = numeric_latent_dim_candidates
 
-        # parameters for a single column of concatenated strings
-        default_hps['concat'] = default_hps['string'].copy()
+        if numeric_hidden_layers_candidates:
+            default_hps['numeric']['numeric_hidden_layers'] = numeric_hidden_layers_candidates
 
         if hps is None:
             hps = {}
 
+        # give parameters in `hps` precedence over default parameters
+        parameters_in_both = set(default_hps.keys()).intersection(set(hps.keys()))
+        for param in parameters_in_both:
+            del default_hps[param]
+        hps = merge_dicts(hps, default_hps)
+
         if user_defined_scores is None:
             user_defined_scores = []
 
@@ -286,22 +291,25 @@ class SimpleImputer:
             train_df, test_df = random_split(train_df, [1-test_split, test_split])
 
         self.check_data_types(train_df)  # infer data types, saved self.string_columns, self.numeric_columns
-        self.hpo.tune(train_df, test_df, hps, strategy, num_evals,
-                      max_running_hours, user_defined_scores, hpo_run_name, self)
+        self.hpo.tune(train_df, test_df, hps, num_evals, max_running_hours, user_defined_scores, hpo_run_name, self)
         self.save()
 
+        return self
+
     def fit(self,
             train_df: pd.DataFrame,
             test_df: pd.DataFrame = None,
             ctx: mx.context = get_context(),
             learning_rate: float = 4e-3,
-            num_epochs: int = 10,
-            patience: int = 3,
+            num_epochs: int = 100,
+            patience: int = 5,
             test_split: float = .1,
             weight_decay: float = 0.,
             batch_size: int = 16,
             final_fc_hidden_units: List[int] = None,
-            calibrate: bool = True) -> Any:
+            calibrate: bool = True,
+            class_weights: dict = None,
+            instance_weights: list = None) -> Any:
         """
 
         Trains and stores imputer model
@@ -320,11 +328,14 @@ class SimpleImputer:
         :param weight_decay: regularizer (default 0)
         :param batch_size (default 16)
         :param final_fc_hidden_units: list dimensions for FC layers after the final concatenation
-
+        :param calibrate: Control automatic model calibration
+        :param class_weights: Dictionary with labels as keys and weights as values.
+                              Weighs each instance's contribution to the likelihood based on the corresponding class.
+        :param instance_weights: List of weights for each instance in train_df.
         """
 
-        if final_fc_hidden_units is None:
-            final_fc_hidden_units = [100]
+        # add weights to training data if provided
+        train_df = self.__add_weights_to_df(train_df, class_weights, instance_weights, in_place=False)
 
         self.check_data_types(train_df)
 
@@ -356,19 +367,19 @@ class SimpleImputer:
                 NumericalFeaturizer(field_name=numerical_feature_column, numeric_latent_dim=self.numeric_latent_dim,
                                     numeric_hidden_layers=self.numeric_hidden_layers)]
 
-        label_column = []
-
         if is_numeric_dtype(train_df[self.output_column]):
             label_column = [NumericalEncoder(self.output_column, normalize=True)]
-            logger.info("Assuming numeric output column: {}".format(self.output_column))
+            logger.debug("Assuming numeric output column: {}".format(self.output_column))
         else:
             label_column = [CategoricalEncoder(self.output_column, max_tokens=self.num_labels)]
-            logger.info("Assuming categorical output column: {}".format(self.output_column))
+            logger.debug("Assuming categorical output column: {}".format(self.output_column))
 
-        self.imputer = Imputer(data_encoders=data_encoders,
-                               data_featurizers=data_columns,
-                               label_encoders=label_column,
-                               output_path=self.output_path)
+        # to make consecutive calls to .fit() continue where the previous call finished
+        if self.imputer is None:
+            self.imputer = Imputer(data_encoders=data_encoders,
+                                   data_featurizers=data_columns,
+                                   label_encoders=label_column,
+                                   output_path=self.output_path)
 
         self.output_path = self.imputer.output_path
 
@@ -447,57 +458,81 @@ class SimpleImputer:
     @staticmethod
     def complete(data_frame: pd.DataFrame,
                  precision_threshold: float = 0.0,
-                 inplace: bool = False):
+                 inplace: bool = False,
+                 hpo: bool = False,
+                 verbose: int = 0,
+                 num_epochs: int = 100,
+                 iterations: int = 1,
+                 output_path: str = "."):
         """
         Given a dataframe with missing values, this function detects all imputable columns, trains an imputation model
         on all other columns and imputes values for each missing value.
-
+        Several imputation iterators can be run.
         Imputable columns are either numeric columns or non-numeric categorical columns; for determining whether a
             column is categorical (as opposed to a plain text column) we use the following heuristic:
             a non-numeric categorical column should have least 10 times as many rows as there were unique values
-
         If an imputation model did not reach the precision specified in the precision_threshold parameter for a given
             imputation value, that value will not be imputed; thus depending on the precision_threshold, the returned
             dataframe can still contain some missing values.
-
         For numeric columns, we do not filter for accuracy.
-
         :param data_frame: original dataframe
         :param precision_threshold: precision threshold for categorical imputations (default: 0.0)
-        :param inplace: whether or not to perform imputations inplace
+        :param inplace: whether or not to perform imputations inplace (default: False)
+        :param hpo: whether or not to perform hyperparameter optimization (default: False)
+        :param verbose: verbosity level, values > 0 log to stdout (default: 0)
+        :param num_epochs: number of epochs for each imputation model training (default: 100)
+        :param iterations: number of iterations for iterative imputation (default: 1)
+        :param output_path: path to store model and metrics
         :return: dataframe with imputations
-
         """
 
         # TODO: should we expose temporary dir for model serialization to avoid crashes due to not-writable dirs?
+        
+        missing_mask = data_frame.copy().isnull()
 
         if inplace is False:
             data_frame = data_frame.copy()
 
+        if verbose == 0:
+            set_stream_log_level("ERROR")
+
         numeric_columns = [c for c in data_frame.columns if is_numeric_dtype(data_frame[c])]
         string_columns = list(set(data_frame.columns) - set(numeric_columns))
-        logger.info("Assuming numerical columns: {}".format(", ".join(numeric_columns)))
+        logger.debug("Assuming numerical columns: {}".format(", ".join(numeric_columns)))
 
         col_set = set(numeric_columns + string_columns)
 
         categorical_columns = [col for col in string_columns if SimpleImputer._is_categorical(data_frame[col])]
-        logger.info("Assuming categorical columns: {}".format(", ".join(categorical_columns)))
+        logger.debug("Assuming categorical columns: {}".format(", ".join(categorical_columns)))
+        for _ in range(iterations):
+            for output_col in set(numeric_columns) | set(categorical_columns):
+                # train on all input columns but the to-be-imputed one
+                input_cols = list(col_set - set([output_col]))
 
-        for output_col in set(numeric_columns) | set(categorical_columns):
-            # train on all input columns but the to-be-imputed one
-            input_cols = list(col_set - set([output_col]))
+                # train on all observed values
+                idx_missing = missing_mask[output_col]
 
-            # train on all observed values
-            idx_missing = data_frame[output_col].isnull()
+                imputer = SimpleImputer(input_columns=input_cols,
+                                        output_column=output_col,
+                                        output_path=os.path.join(output_path, output_col))
+                if hpo:
+                    imputer.fit_hpo(data_frame.loc[~idx_missing, :],
+                                    patience=5 if output_col in categorical_columns else 20,
+                                    num_epochs=100,
+                                    final_fc_hidden_units=[[0], [10], [50], [100]])
+                else:
+                    imputer.fit(data_frame.loc[~idx_missing, :],
+                                patience=5 if output_col in categorical_columns else 20,
+                                num_epochs=num_epochs,
+                                calibrate=False)
 
-            imputer = SimpleImputer(input_columns=input_cols,
-                                    output_column=output_col) \
-                .fit(data_frame.loc[~idx_missing, :],
-                     patience=5 if output_col in categorical_columns else 20)
-
-            tmp = imputer.predict(data_frame, precision_threshold=precision_threshold)
-            data_frame.loc[idx_missing, output_col] = tmp[output_col + "_imputed"]
+                tmp = imputer.predict(data_frame, precision_threshold=precision_threshold)
+                data_frame.loc[idx_missing, output_col] = tmp[output_col + "_imputed"]
 
+                # remove the directory with logfiles for this column 
+                shutil.rmtree(os.path.join(output_path, output_col))
+                
+                
         return data_frame
 
     def save(self):
@@ -532,7 +567,7 @@ class SimpleImputer:
 
         """
 
-        logger.info("Output path for loading Imputer {}".format(output_path))
+        logger.debug("Output path for loading Imputer {}".format(output_path))
         # load pickled model
         simple_imputer_params = pickle.load(
             open(os.path.join(output_path, "simple_imputer.pickle"), "rb"))
@@ -567,15 +602,16 @@ class SimpleImputer:
         """
 
         if self.hpo.results is None:
-            logger.warn('No hpo run available. Run hpo calling SimpleImputer.fit_hpo().')
+            logger.warning('No hpo run available. Run hpo by calling SimpleImputer.fit_hpo().')
             return
 
         if hpo_name is None:
             if self.output_type == 'numeric':
-                hpo_name = self.hpo.results['mse'].astype(float).idxmax()
+                hpo_name = self.hpo.results['mse'].astype(float).idxmin()
+                logger.debug("Selecting imputer with minimal mean squared error.")
             else:
                 hpo_name = self.hpo.results['precision_weighted'].astype(float).idxmax()
-            logger.info("Selecting imputer with maximum weighted precision.")
+                logger.debug("Selecting imputer with maximal weighted precision.")
 
         # copy artifacts from hp run to self.output_path
         imputer_dir = self.output_path + str(hpo_name)
@@ -587,3 +623,107 @@ class SimpleImputer:
 
         self.imputer = Imputer.load(self.output_path)
 
+    def __deserialize_confusion_matrix(self) -> np.ndarray:
+        """
+        Return normalized confusion matrix for trained simple_imputer
+        """
+
+        # labels need to be sorted consistently for computation of confusion matrices etc.
+        labels = sorted(self.imputer.label_encoders[0].idx_to_token.values())
+
+        # invert normalized empirical confusion matrix for test data
+        ids = dict((label, index) for (index, label) in enumerate(labels))
+        confusion_matrix_from_file = self.load_metrics()['confusion_matrix']
+        confusion_test = np.empty([len(confusion_matrix_from_file), len(confusion_matrix_from_file)])
+
+        # make matrix from nested dictionary that stores the confusion matrix
+        for row in confusion_matrix_from_file:
+            for entry in row:
+                if type(entry) is str:
+                    row_idx = ids[entry]
+                else:
+                    for col in entry:
+                        col_idx = ids[col[0]]
+                        confusion_test[row_idx, col_idx] = col[1]
+
+        return confusion_test / confusion_test.sum()  # normalize
+
+    def check_for_label_shift(self, target_data: pd.DataFrame) -> dict:
+
+        """
+        Detect label shift in the validation data
+
+        :param test_data: data frame that contains labels
+        :param target_data: unlabelled data for which predictions are to be generated
+
+        :return: dictionary with labels as keys and weights as values.
+        """
+
+        # labels need to be sorted consistently for computation of confusion matrices etc.
+        labels = sorted(self.imputer.label_encoders[0].idx_to_token.values())
+
+        imputed_target = self.predict(target_data)  # generate predictions for test and target data
+        confusion_test = self.__deserialize_confusion_matrix()
+
+        # compute estimates of label marginals for test and target data
+        marginals_test = pd.Series(confusion_test.sum(axis=0), index=labels)
+        marginals_target = imputed_target[self.output_column + '_imputed'].value_counts(
+            normalize=True, sort=False)[labels]
+
+        # estimate the ratio of marginals and store as dictionary.
+        label_weights = np.linalg.solve(confusion_test, marginals_target)
+        label_weights_dict = dict((label, max(weight, 0)) for label, weight in zip(labels, label_weights))
+
+        # estimate marginals of true labels
+        true_marginals_target = np.diag(marginals_test).dot(label_weights)
+
+        logger.warning('\n\tThe estimated label marginals are ' + str(list(zip(labels, true_marginals_target))) +
+                    '\n\tMarginals in the training data are ' + str(list(zip(labels, marginals_test))) +
+                    '\n\tReweighing factors for empirical risk minimization' + str(label_weights_dict))
+
+        if np.any(marginals_test < 0):
+            logger.warning('\n\tEstimated label marginals are invalid. Proceed with caution.')
+
+        return label_weights_dict
+
+    def __add_weights_to_df(self,
+                            df: pd.DataFrame,
+                            class_weights: dict = None,
+                            instance_weights: list = None,
+                            in_place: bool = True) -> pd.DataFrame:
+        """
+        Add additional column to data frame inplace, with entries provided either
+        (1) as dict values in class_dictionary with rows selected on dict keys in weights dictionary
+            in correspondence to label_column.
+        (2) or as list with weights for each instance
+
+        :param df: input data
+        :param class_weights: dictionary with label names and corresponding weights
+        :param instance_weights: list of weights for each instance
+        :param label_column: name of label column
+        :param in_place: If true, append column to df, otherwise create copy.
+        """
+
+        # Nothing to do if no weights have been passed.
+        if class_weights is None and instance_weights is None:
+            return df
+
+        # Check that only one type of weights is provided
+        assert (class_weights is None or instance_weights is None), \
+            "Please provide class_weights XOR instance_weights."
+
+        if in_place is False:
+            df_new = df.copy()
+        else:
+            df_new = df
+
+        if class_weights is not None:
+            df_new[INSTANCE_WEIGHT_COLUMN] = 0.0
+            for label, weight in class_weights.items():
+                df_new.loc[df_new[self.output_column] == label, INSTANCE_WEIGHT_COLUMN] = weight
+
+        elif instance_weights is not None:
+            df_new[INSTANCE_WEIGHT_COLUMN] = instance_weights
+
+        return df_new
+
