@@ -17,25 +17,24 @@ DataWig imputer tests
 
 """
 
-import os
-import random
-import warnings
-
 import numpy as np
+import os
 import pandas as pd
 import pytest
+import warnings
+from sklearn.metrics import precision_score
+from stat import *
+import string
 
+import datawig
 from datawig.column_encoders import (BowEncoder, CategoricalEncoder,
                                      NumericalEncoder, SequentialEncoder,
                                      TfIdfEncoder)
-from datawig.imputer import Imputer
+from datawig.imputer import (Imputer, INSTANCE_WEIGHT_COLUMN)
 from datawig.mxnet_input_symbols import (BowFeaturizer, EmbeddingFeaturizer,
                                          LSTMFeaturizer, NumericalFeaturizer)
 from datawig.utils import random_split
-import os
-import datawig
-from datawig.utils import random_split
-from sklearn.metrics import precision_score
+
 
 warnings.filterwarnings("ignore")
 
@@ -360,12 +359,12 @@ def test_imputer_without_train_df(test_dir):
         output_path=output_path,
     )
 
-    with pytest.raises(ValueError, message="Need a non-empty DataFrame for fitting Imputer model"):
+    with pytest.raises(ValueError, match="Need a non-empty DataFrame for fitting Imputer model"):
         imputer.fit(
             train_df=df_train
         )
 
-    with pytest.raises(ValueError, message="Need a non-empty DataFrame for fitting Imputer model"):
+    with pytest.raises(ValueError, match="Need a non-empty DataFrame for fitting Imputer model"):
         imputer.fit(
             train_df=None
         )
@@ -456,6 +455,29 @@ def test_imputer_load_read_exec_only_dir(tmpdir, data_frame):
         print(e)
         pytest.fail('Loading imputer from read-only directory should not fail.')
 
+def test_imputer_load_with_invalid_context(tmpdir, data_frame):
+    
+    # on shared build-fleet tests fail with converting tmpdir to string
+    tmpdir = str(tmpdir)
+    feature = 'feature'
+    label = 'label'
+
+    df = data_frame(feature, label, n_samples=100)
+    # fit and output model + metrics to tmpdir
+
+    imputer = Imputer(
+        data_featurizers=[BowFeaturizer(feature)],
+        label_encoders=[CategoricalEncoder(label)],
+        data_encoders=[BowEncoder(feature)],
+        output_path=tmpdir
+    )
+    imputer.fit(train_df=df, num_epochs=1)
+    imputer.ctx = None
+    imputer.save()
+
+    imputer_deser = Imputer.load(tmpdir)
+    _ = imputer_deser.predict(df)
+    
 
 def test_imputer_fit_fail_non_writable_output_dir(tmpdir, data_frame):
     import stat
@@ -599,7 +621,7 @@ def test_mxnet_module_wrapper(data_frame):
     mod = _MXNetModule(mx.current_context(), label_encoders, data_featurizers, final_fc_hidden_units=[])(iter_train)
 
     assert mod._label_names == [label_col]
-    assert mod.data_names == [feature_col]
+    assert sorted(mod.data_names) == sorted([feature_col] + [INSTANCE_WEIGHT_COLUMN])
     # weights and biases
     assert len(mod._arg_params) == 2
 
@@ -745,3 +767,76 @@ def test_explain_method_synthetic(test_dir):
     imputer.save()
     imputer_from_disk = Imputer.load(imputer.output_path)
     assert np.all(['f' in token for token, weight in imputer_from_disk.explain('foo')['in_text']][:3])
+
+
+def test_non_writable_output_path(test_dir, data_frame):
+    label_col = 'label'
+    df = data_frame(n_samples=100, label_col=label_col)
+
+    data_encoder_cols = [TfIdfEncoder('features')]
+    label_encoder_cols = [CategoricalEncoder(label_col)]
+    data_cols = [BowFeaturizer('features')]
+
+    output_path = os.path.join(test_dir, 'non_writable')
+
+    Imputer(
+        data_featurizers=data_cols,
+        label_encoders=label_encoder_cols,
+        data_encoders=data_encoder_cols,
+        output_path=output_path
+    ).fit(
+        train_df=df,
+        num_epochs=1
+    ).save()
+
+    from datawig.utils import logger
+
+    try:
+        # make output dir of imputer read-only
+        os.chmod(output_path, S_IREAD | S_IXUSR)
+
+        # make log file read only
+        os.chmod(os.path.join(output_path, "imputer.log"), S_IREAD)
+        imputer = Imputer.load(output_path)
+        _ = imputer.predict(df)
+        logger.warning("this should not fail")
+
+        # remove log file
+        os.chmod(os.path.join(output_path, "imputer.log"), S_IREAD | S_IXUSR | S_IWUSR)
+        os.chmod(output_path, S_IREAD | S_IXUSR | S_IWUSR)
+        os.remove(os.path.join(output_path, "imputer.log"))
+
+        # make output dir of imputer read-only
+        os.chmod(output_path, S_IREAD | S_IXUSR)
+
+        imputer = Imputer.load(output_path)
+        _ = imputer.predict(df)
+        logger.warning("this should not fail")
+        os.chmod(output_path, S_IREAD | S_IXUSR | S_IWUSR)
+    except Exception as e:
+        print(e)
+        pytest.fail("This invocation not raise any Exception")
+
+
+def test_fit_resumes(test_dir, data_frame):
+    feature_col, label_col = "feature", "label"
+
+    df = data_frame(feature_col=feature_col,
+                    label_col=label_col)
+
+    imputer = Imputer(
+        data_encoders=[TfIdfEncoder([feature_col])],
+        data_featurizers=[datawig.mxnet_input_symbols.BowFeaturizer(feature_col)],
+        label_encoders=[CategoricalEncoder(label_col)],
+        output_path=test_dir
+    )
+
+    assert imputer.module is None
+
+    imputer.fit(df, num_epochs=20)
+    first_fit_module = imputer.module
+
+    imputer.fit(df, num_epochs=20)
+    second_fit_module = imputer.module
+
+    assert first_fit_module == second_fit_module
